/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <gnome.h>

#include <sys/types.h>
#include <ctype.h>

#include "date.h"
#include "globals.h"
#include "gnksa.h"
#include "grouplist.h"
#include "log.h"
#include "message.h"
#include "message-send.h"
#include "message-window.h"
#include "nntp.h"
#include "queue.h"
#include "task-send.h"
#include "smtp.h"
#include "util.h"

static gchar* generate_references (Message *msg);

/***
****  PRIVATE UTILITIES
***/

/* Find and return a list of e-mail addresses to send to */
static GPtrArray *
get_email_rcpts (const gchar *to)
{
	GPtrArray * a = g_ptr_array_new ();
	GString * rcpt = g_string_new (NULL);
	gchar * tmp;
	const gchar * march;

	/* make a temporary copy and make all the delimiters into spaces */
	tmp = g_strdup (to);
	g_strdelimit (tmp, ";:", ',');

	/* march through the tokens & save the nonempty ones */
	march = tmp;
	while (get_next_token_g_str (march, ',', &march, rcpt)) {
		pan_g_string_strstrip (rcpt);
		g_message ("[%s]", rcpt->str);
		if (rcpt->len != 0)
			g_ptr_array_add (a, g_strdup(rcpt->str));
	}

	/* cleanup */
	g_free (tmp);
	g_string_free (rcpt, TRUE);

	return a;
}

/* Find and return a list of NNTP groups to send to */
static GPtrArray*
get_nntp_rcpts (const gchar *to)
{
	GPtrArray * a = g_ptr_array_new ();
	GString * rcpt = g_string_new (NULL);
	gchar * tmp;
	const gchar * march;

	/* make a temporary copy and make all the delimiters into spaces */
	tmp = g_strdup (to);
	g_strdelimit (tmp, " ;:", ',');

	/* march through the tokens & save the nonempty ones */
	march = tmp;
	while (get_next_token_g_str (march, ',', &march, rcpt)) {
		pan_g_string_strstrip (rcpt);
		if (rcpt->len != 0)
			g_ptr_array_add (a, g_strdup(rcpt->str));
	}

	/* cleanup */
	g_free (tmp);
	g_string_free (rcpt, TRUE);

	return a;
}


/***
****  OUTGOING MESSAGE CHECKS
***/

typedef enum
{
	OKAY,
	WARN,
	REFUSE,
}
GoodnessLevel;


/**
 * Check to see if the signature (if found) is within the McQuary limit of
 * four lines and 80 columns per line.
 */
static void
check_signature (GPtrArray * setme_errors,
                 GoodnessLevel * setme_goodness,
                 const char * body,
                 GString * line)
{
	int too_wide_qty = 0;
	int sig_line_qty = -1;

	g_return_if_fail (setme_errors!=NULL);
	g_return_if_fail (setme_goodness!=NULL);
	g_return_if_fail (body!=NULL);
	g_return_if_fail (line!=NULL);

	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line))
	{
		if (strcmp(line->str,"-- ") && sig_line_qty<0)
			continue;

		++sig_line_qty;

		if (line->len>80)
			++too_wide_qty;
	}

	if (sig_line_qty == 0)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Signature prefix with no signature.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}

	else if (sig_line_qty > 4)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: signature is more than 4 lines long")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}

	if (too_wide_qty != 0)
	{
		g_ptr_array_add (setme_errors, g_strdup_printf(
			_("WARNING: %d lines too wide in signature"),
			too_wide_qty));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}


/**
 * Simple check to see if the body is too wide.  Any text after the
 * signature prefix is ignored in this test.
 */
static void
check_wide_body (GPtrArray * setme_errors,
                 GoodnessLevel * setme_goodness,
                 const gchar * body,
		 GString * line)
{
	int too_wide_qty = 0;

	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line))
	{
		if (!strcmp (line->str, "-- "))
			break;
		if (line->len>80)
			++too_wide_qty;
	}

	if (too_wide_qty != 0)
	{
		g_ptr_array_add (setme_errors, g_strdup_printf(
			_("WARNING: %d lines >80 characters in the message."),
			too_wide_qty));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}
/**
 * Check to see if the article appears to be empty.  Any text after the
 * signature prefix is ignored in this test.
 */
static void
check_empty (GPtrArray * setme_errors,
             GoodnessLevel * setme_goodness,
             const gchar* body,
             GString * line)
{
	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line))
	{
		if (!strcmp (line->str, "-- "))
			break;
		pan_g_string_strstrip (line);
		if (line->len != 0)
			return;
	}

	g_ptr_array_add (setme_errors, g_strdup(
		_("ERROR: Message is empty.")));
	*setme_goodness = MAX (*setme_goodness, REFUSE);
}

/**
 * Check to see how much original content is in this message, opposed
 * to quoted content.  Any text after the signature prefix is ignored
 * in this test.
 *
 * (1) count all the lines beginning with the quoted prefix.
 * (2) count all the nonempty nonsignature lines.  These are the orignal lines.
 * (3) if the ratio of original/quoted is 20% or less, warn.
 * (4) if the ratio of original/quoted is 0%, warn louder.
 */
static void
check_mostly_quoted (GPtrArray * setme_errors,
                     GoodnessLevel * setme_goodness,
                     const gchar * body,
                     GString * line)
{
	int unquoted = 0;
	int total = 0;

	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line)) {
		if (!strcmp (line->str, "-- "))
			break;
		pan_g_string_strstrip (line);
		if (line->len==0)
			continue;
		++total;
		if (*line->str != '>')
			++unquoted;
	}

	if (total!=0 && ((unquoted/(double)total)<0.2))
	{
		char * pch = g_strdup (unquoted==0
			?  _("WARNING: The message is entirely quoted text!")
			:  _("WARNING: The message is mostly quoted text."));
		g_ptr_array_add (setme_errors, pch);
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}

/**
 * Check to see if the article appears to only have quoted text.  If this
 * appears to be the case, we will refuse to post the message.
 *
 * (1) Determine what the attribution string should be.
 *
 * (2) Replace carriage returns in both the calculated attribution string
 *     and a temporary copy of the message body, so that we don't have to
 *     worry whether or not the attribution line's been wrapped.
 * 
 * (3) Search for an occurance of the attribution string in the body.  If
 *     it's found, remove it from the temporary copy of the body so that
 *     it won't affect our line counts.
 *
 * (4) Of the remaining body, look for any nonempty lines before the signature
 *     file that don't begin with the quote prefix.  If such a line is found,
 *     then the message is considered to not be all quoted text.
 *
 */
static void
check_all_quoted (GPtrArray * setme_errors,
                  GoodnessLevel * setme_goodness,
	          const gchar* body,
                  GString * line,
		  const Message * msg)
{
	int attribution_len;
	int attribution_start_index;
	char * body2;
	const char * pch;
	GString * attribution;

	/* (1) */
       	attribution = g_string_new (NULL);
	message_get_attribution_string (msg, attribution);
	body2 = g_strdup (body);
	attribution_len = attribution->len;

	/* (2) */
	g_strdelimit (attribution->str, "\n", ' ');
	g_strdelimit (body2, "\n", ' ');

	/* (3) */
	pch = strstr (body2, attribution->str);
	attribution_len = attribution->len;
	attribution_start_index = pch==NULL ? -1 : pch-body2;
	g_free (body2);
	body2 = g_strdup (body);
	g_string_free (attribution, TRUE);
	if (attribution_start_index != -1) {
		char * src = body2 + attribution_start_index + attribution_len;
		char * tgt = body2 + attribution_start_index;
		while (*src)
			*tgt++ = *src++;
		*tgt = '\0';
	}

	/* (4) */
	pch = body2;
	while (get_next_token_g_str (pch, '\n', &pch, line)) {
		if (!strcmp (line->str, "-- "))
			break;
		pan_g_string_strstrip (line);
		if (line->len==0)
			continue;
		if (*line->str != '>') {
			g_free (body2);
			return;
		}
	}
	g_free (body2);

	g_ptr_array_add (setme_errors, g_strdup(
		_("ERROR: Message appears to have no new content.")));
	*setme_goodness = MAX (*setme_goodness, REFUSE);
}

static void
check_body (GPtrArray * setme_errors,
	    GoodnessLevel *setme_goodness,
	    const gchar * body,
	    const Message * msg)
{
	GString * line;

	g_return_if_fail (setme_goodness != NULL);
	g_return_if_fail (setme_errors != NULL);
	g_return_if_fail (body != NULL);
	g_return_if_fail (msg != NULL);

	line = g_string_sized_new (1024);
	check_empty (setme_errors, setme_goodness, body, line);
	check_all_quoted (setme_errors, setme_goodness, body, line, msg);
	check_wide_body (setme_errors, setme_goodness, body, line);
	check_signature (setme_errors, setme_goodness, body, line);
	check_mostly_quoted (setme_errors, setme_goodness, body, line);
	g_string_free (line, TRUE);
}




/**
 * @param msg the message being replied to.
 * @return a newly-allocated references string.
 */
static gchar*
generate_references (Message *msg)
{
	GString * str;
	gchar * retval;

	/* sanity check */
	g_return_val_if_fail (msg!=NULL, NULL);

	/* build a full references string */
       	str = g_string_sized_new (1024);
	if (is_nonempty_string(msg->references)) {
		g_string_append (str, msg->references);
		g_string_append_c (str, ' ');
	}
	g_string_append (str, msg->message_id);

	/* make sure they fit gnksa guidelines */
	retval = gnksa_trim_references (str->str);
	g_string_free (str, TRUE);

	return retval;
}

/**
 * thus spake son-of-1036: "the most popular method of generating local parts
 * is to use the date and time, plus some way of distinguishing between
 * simultaneous postings on the same host (e.g., a process number), and encode
 * them in a suitably-reduced alphabet.
 */
static gchar*
generate_message_id (void)
{
	char * tmp;
	GString * msgid = g_string_sized_new (256);

	/* start with '<' */
	g_string_append_c (msgid, '<');

	/* add local part to message-id */
	{
		char buf[32];
		const time_t now = time(NULL);
		struct tm local;
		localtime_r (&now, &local);
		strftime (buf, sizeof(buf), "%Y%m%d.%H%M%S", &local);
		g_string_append (msgid, buf);
		g_string_sprintfa (msgid, ".%lu", (gulong)getpid());
	}

	/* delimit */
	g_string_append_c (msgid, '@');

	/* add fqdn to message-id */
	{
		gchar * host = get_host_name ();
		gchar * fqdn = get_fqdn (host);
		g_string_append (msgid, fqdn);
		g_free (host);
		g_free (fqdn);
	}

	/* end with '>' */
	g_string_append_c (msgid, '>');

	/* return */
	tmp = msgid->str;
	g_string_free (msgid, FALSE);
	return tmp;
}


static void
check_followup_to (GPtrArray        * setme_errors,
		   GoodnessLevel    * setme_goodness,
		   const Server     * server,
		   const GPtrArray  * group_names)
{
	int i;

	g_return_if_fail (setme_errors != NULL);
	g_return_if_fail (setme_goodness != NULL);
	g_return_if_fail (server != NULL);
	g_return_if_fail (group_names != NULL);

	/* check to make sure all the groups exist */
	for (i=0; i<group_names->len; ++i)
	{
		const char * name = (char*) g_ptr_array_index (group_names, i);
		if (g_strcasecmp(name, "poster")
			&& !server_has_group (server, name))
		{
			g_ptr_array_add (setme_errors, g_strdup_printf(
				_("WARNING: Unknown group: \"%s\"."), name));
			*setme_goodness = MAX (*setme_goodness, WARN);
		}
	}

	/* warn if too many followup-to groups */
	if (group_names->len > 5)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Following-Up to many groups.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}

static void
check_subject (GPtrArray * setme_errors,
	       GoodnessLevel * setme_goodness,
	       const gchar * subject)
{
	g_assert (setme_errors!=NULL);
	g_assert (setme_goodness!=NULL);

	if (!is_nonempty_string (subject))
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("ERROR: No subject specified.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}
}

static void
check_groups (GPtrArray          * setme_errors,
	      GoodnessLevel      * setme_goodness,
	      const Server       * server,
	      const GPtrArray    * group_names,
	      gboolean             followup_to_set)
{
	int i;

	g_return_if_fail (setme_errors!=NULL);
	g_return_if_fail (setme_goodness!=NULL);
	g_return_if_fail (server!=NULL);
	g_return_if_fail (group_names!=NULL);

	/* check to make sure all the groups exist */
	for (i=0; i<group_names->len; ++i)
	{
		const char * name = (char*) g_ptr_array_index (group_names, i);
		if (!server_has_group (server, name))
		{
			g_ptr_array_add (setme_errors, g_strdup_printf(
				_("WARNING: Unknown group: \"%s\"."), name));
			*setme_goodness = MAX (*setme_goodness, WARN);
		}
	}

	/* refuse if far too many groups */
	if (group_names->len > 12)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("ERROR: Posting to a very large number of groups.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}
	/* warn if too many groups */
	else if (group_names->len > 5)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Posting to a large number of groups.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}


	/* warn if too many groups and no followup-to */
	if (group_names->len>2 && !followup_to_set)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Crossposting without setting Followup-To.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}

/**
 * @return the number of recipients
 */
static int
check_mail_addresses (GPtrArray * setme_errors,
		      GoodnessLevel * setme_goodness,
		      const char * addresses_str)
{
	GPtrArray * a = get_email_rcpts (addresses_str);
	int i;

	for (i=0; i<a->len; ++i)
	{
		gchar * from = (gchar*) g_ptr_array_index (a, i);
		int val = gnksa_check_from (from);
		if (val != GNKSA_OK)
		{
			*setme_goodness = MAX (*setme_goodness, REFUSE);
			g_ptr_array_add (setme_errors, g_strdup_printf (
				_("ERROR: %s ``%s''"),
				gnksa_status_text(val), from));
		}
	}

	/* cleanup */
	i = a->len;
	pan_g_ptr_array_foreach (a, (GFunc)g_free, NULL);
	g_ptr_array_free (a, TRUE);
	return i;
}



/**
 * Note that error-checking is done elsewhere.
 * By the time we reach this point the message is
 * assumed to be clear to post 
 */
static gchar*
build_nntp_message (Message *msg)
{
	/*static const char * fqdn = NULL;*/
	static const char * host = NULL;
	static char * uname = NULL;
	gchar * tmp;
	int i;
	GString * body = g_string_sized_new (1024);
	GString * postme = g_string_sized_new (1024);

	/* static setup... */
	if (host == NULL)
		host = get_host_name ();
	/*if (fqdn==NULL && host!=NULL)
		fqdn = get_fqdn (host);*/
	if (uname == NULL)
		uname = g_get_user_name ();

	/* sanity checks... */
	/*g_return_val_if_fail (fqdn!=NULL, NULL);*/
	g_return_val_if_fail (msg!=NULL, NULL);
	g_return_val_if_fail (host!=NULL, NULL);
	g_return_val_if_fail (uname!=NULL, NULL);
	g_return_val_if_fail (msg->rcpts!=NULL, NULL);

	/**
	***  Build the Article
	**/

	/* From: */
	tmp = message_window_get_from (msg->window);
	g_string_sprintfa (postme, "From: %s\r\n", tmp);
	g_free (tmp);

	/* Subject: */
	tmp = message_window_get_subject (msg->window);
	g_string_sprintfa (postme, "Subject: %s\r\n", tmp);
	g_free (tmp);

	/* Date: */
	tmp = generate_date ();
	g_string_sprintfa (postme, "Date: %s\r\n", tmp);
	g_free (tmp);

	/* Newsgroups: */
	g_string_append (postme, "Newsgroups: ");
	for (i=0; i<msg->rcpts->len; ++i)
	{
		char * name = (char*) g_ptr_array_index (msg->rcpts, i);
		g_string_append (postme, name);
		if (i < msg->rcpts->len-1)
			g_string_append_c (postme, ',');
	}
	g_string_append (postme, "\r\n");

	/* Path: */
	g_string_sprintfa (postme, "Path: %s!%s\r\n", host, uname);

	/* Followup-To: */
	tmp = message_window_get_followup_to (msg->window);
	if (tmp != NULL)
	{
		GPtrArray * a = get_nntp_rcpts (tmp);

		if (a->len!=0)
		{
			int i;
			g_string_sprintfa (postme, "Followup-To: ");
			for (i=0; i<a->len; ++i)
			{
				char * name = (char*) g_ptr_array_index (a, i);
				g_string_append (postme, name);
				if (i<a->len-1)
					g_string_append_c (postme, ',');
			}
			g_string_sprintfa (postme, "\r\n");
		}

		pan_g_ptr_array_foreach (a, (GFunc)g_free, NULL);
		g_ptr_array_free (a, TRUE);
		g_free (tmp);
	}

	/* Reply-To: */
	tmp = message_window_get_reply_to (msg->window);
	if (is_nonempty_string(tmp))
		g_string_sprintfa (postme, "Reply-To: %s\r\n", tmp);
	g_free (tmp);

	/* Organization: */
	tmp = message_window_get_organization (msg->window);
	if (is_nonempty_string(tmp))
		g_string_sprintfa (postme, "Organization: %s\r\n", tmp);
	g_free (tmp);

	/* Message ID: (if desired) ... */
	if (msg->server->gen_msgid)
	{
		tmp = generate_message_id ();
		g_string_sprintfa (postme, "Message-ID: %s\r\n", tmp);
		g_free (tmp);
	}

	/* References: */
	if (msg->type==NNTP_REPLY)
	{
		tmp = generate_references (msg);
		if (is_nonempty_string (tmp))
			g_string_sprintfa (postme, "References: %s\r\n", tmp);
		g_free (tmp);
	}

	/* User Agent: */
	g_string_append (postme, "User-Agent: Pan/" VERSION " (Unix)\r\n");

	/* separator that goes between headers and body */
	g_string_sprintfa (postme, "\r\n");

	/* body */
	tmp = message_window_get_body (msg->window);
	g_string_assign (body, tmp);
	g_free (tmp);
	pan_g_string_replace (body, "\n.", "\n.."); /* RFC977 3.10.1 ... */
	pan_g_string_replace (body, "\n", "\r\n"); /* ..make sure that each.. */
	pan_g_string_replace (body, "\r\r\n", "\r\n"); /* ..line ends in \r\n */
	g_string_append (postme, body->str);

	/* cleanup */
	tmp = postme->str;
	g_string_free (postme, FALSE);
	g_string_free (body, TRUE);
	return tmp;
}



static int
send_done_cb (gpointer call_object, gpointer call_arg, gpointer user_data)
{
	const int status = GPOINTER_TO_INT(call_arg);
	StatusItem * item = STATUS_ITEM(call_object);
	GtkWidget * send_button = GTK_WIDGET(user_data);
	Message * message = TASK_SEND(item)->msg;
	GString * text = g_string_new (NULL);

	if (status & SMTP_FAILED)
		g_string_append (text, _("E-Mail send failed.\n"));
	if (status & NNTP_FAILED)
		g_string_append (text, _("Usenet posting failed.\n"));
	if (!status)
		g_string_append (text, _("Message sent successfully.\n"));

	pan_lock();
	if (text->len != 0)
	{
		/* display status message */
		gnome_ok_dialog_parented (text->str, GTK_WINDOW(Pan.window));
		g_message (text->str);

	}
	if (!status)
	{
		gtk_widget_destroy (message->window->window);
		message_free (message);
	}
	else
	{
		gtk_widget_set_sensitive (send_button, TRUE);
	}
	pan_unlock();

	/* cleanup */
	g_string_free (text, TRUE);
	return 0;
}

static void
start_sending (GtkWidget *widget, Server* server, Message *msg)
{
	PanObject *item = task_send_new (server, msg);
	gtk_widget_set_sensitive (GTK_WIDGET(widget), FALSE);
	pan_callback_add (STATUS_ITEM(item)->done, send_done_cb, widget);
	queue_add (TASK(item));
}

typedef struct
{
	GtkWidget * widget;
	Server * server;
	Message * msg;
}
SendAnywayCallbackStruct;

static void
send_anyway_cb (gint reply, gpointer data)
{
	if (!reply) { // yes
		SendAnywayCallbackStruct *cbs = (SendAnywayCallbackStruct*)data;
		start_sending (cbs->widget, cbs->server, cbs->msg);
	}

	g_free (data);
}

void
send_cb (GtkWidget *widget, Message *msg)
{
	GPtrArray * errors;
	gchar * tmp;
	GoodnessLevel goodness;
	Server * server = grouplist_get_current_server();
	int group_qty;
	int mail_qty;
	int author_qty;
	gboolean followup_to_set;

	/* sanity check */
	g_return_if_fail (widget!=NULL);
	g_return_if_fail (msg!=NULL);
	g_return_if_fail (msg->window!=NULL);
	g_return_if_fail (msg->window->subject!=NULL);

	goodness = OKAY;
       	errors = g_ptr_array_new ();

	/* check the subject... */
	tmp = message_window_get_subject (msg->window);
	check_subject (errors, &goodness, tmp);
	g_free (tmp);

	/* check the body... */
	tmp = message_window_get_body (msg->window);
	check_body (errors, &goodness, tmp, msg);
	g_free (tmp);

	/* check the optional followup-to... */
	tmp = message_window_get_followup_to (msg->window);
	followup_to_set = FALSE;
	if (is_nonempty_string (tmp))
	{
		GPtrArray * a = get_nntp_rcpts (tmp);

		followup_to_set = a->len > 0;
		check_followup_to (errors, &goodness, server, a);

		pan_g_ptr_array_foreach (a, (GFunc)g_free, NULL);
		g_ptr_array_free (a, TRUE);
	}
	g_free (tmp);

	/* check the groups... */
	group_qty = 0;
	tmp = message_window_get_newsgroups (msg->window);
	if (is_nonempty_string(tmp))
	{
		GPtrArray * a = get_nntp_rcpts (tmp);

		group_qty = a->len;
		check_groups (errors, &goodness, server, a, followup_to_set);

		pan_g_ptr_array_foreach (a, (GFunc)g_free, NULL);
		g_ptr_array_free (a, TRUE);
	}
	g_free (tmp);

	/* check the author... */
	tmp = message_window_get_from (msg->window);
	author_qty = check_mail_addresses (errors, &goodness, tmp);
	g_free (tmp);
	if (author_qty!=1) {
		g_ptr_array_add (errors, g_strdup(
			_("ERROR: Must have one author in From: line")));
		goodness = MAX (goodness, REFUSE);
	}

	/* check the email recipients... */
	tmp = message_window_get_users (msg->window);
	mail_qty = check_mail_addresses (errors, &goodness, tmp);
	g_free (tmp);

	/* one last error check */
	if (!group_qty && !mail_qty) {
		g_ptr_array_add (errors, g_strdup(
			_("ERROR: No recipients specified.")));
		goodness = MAX (goodness, REFUSE);
	}

	/* post the warnings, if we have any */
	if (errors->len != 0)
	{
		int i;
		GString * str = g_string_new (NULL);

		for (i=0; i<errors->len; ++i)
		{
			char * error = (char*) g_ptr_array_index (errors, i);
			g_string_append (str, error);
			g_string_append_c (str, '\n');
		}

		if (goodness==REFUSE)
		{
			pan_error_dialog (str->str);
		}
		else
		{
			SendAnywayCallbackStruct* cbs = g_new0 (SendAnywayCallbackStruct, 1);
			const gchar *ques = errors->len > 1
				? _("Do you want to post anyway, despite these warnings?")
				: _("Do you want to post anyway, despite this warning?");
			g_string_append (str, "\n");
			g_string_append (str, ques);
			cbs->widget = widget;
			cbs->server = server;
			cbs->msg = msg;
			gnome_question_dialog (str->str, send_anyway_cb, cbs);
		}
		g_warning (str->str);

		/* cleanup */
		g_string_free (str, TRUE);
		pan_g_ptr_array_foreach (errors, (GFunc)g_free, NULL);
	}

	if (goodness==OKAY)
		start_sending (widget, server, msg);

	g_ptr_array_free (errors, TRUE);
}

gint
message_send (StatusItem * item,
              PanSocket * sock,
              Message * msg)
{
	GPtrArray * mail_rcpts;
	GPtrArray * nntp_rcpts;
	gchar * tmp;
	gint retval = 0;

	/* entry assertions */
	g_return_val_if_fail (msg!=NULL, -1);

	/* who is this being mailed to... */
	tmp = message_window_get_users (msg->window);
	mail_rcpts = get_email_rcpts (tmp);
	g_free (tmp);

	/* where is this being posted... */
	tmp = message_window_get_newsgroups (msg->window);
	nntp_rcpts = get_nntp_rcpts (tmp);
	g_free (tmp);

	/* mail the message */
	if (mail_rcpts->len != 0)
	{
		gchar * leader = NULL;
		gchar * subject = message_window_get_subject (msg->window);
		int val;

		if (nntp_rcpts->len != 0) {
			leader = g_strdup_printf (
				_("[This is an email copy of a Usenet posting to \"%s\"]"),
				(gchar*)g_ptr_array_index(nntp_rcpts,0));
		}

		msg->rcpts = mail_rcpts;
		val = smtp_send (item, msg, leader);
		msg->rcpts = NULL;

		/* mail the message */
		if (!val)
			log_add_va (_("E-mail Message \"%s\" sent."), subject);
		else {
			log_add_va (_("E-mail Message \"%s\" not sent."), subject);
			retval |= SMTP_FAILED;
		}

		/* cleanup */
		g_free (leader);
		g_free (subject);
	}

	/* post the message */
	if (nntp_rcpts->len != 0)
	{
		gchar * nntp_msg;
		gchar * subject = message_window_get_subject (msg->window);

		msg->rcpts = nntp_rcpts;
	       	nntp_msg = build_nntp_message (msg);
		msg->rcpts = NULL;

		/* post the message */
		if (!nntp_post (sock, nntp_msg))
			log_add_va (_("NNTP Article \"%s\" posted."), subject);
		else
			retval |= NNTP_FAILED;

		/* cleanup */
		g_free (nntp_msg);
		g_free (subject);
	}

	pan_g_ptr_array_foreach (nntp_rcpts, (GFunc)g_free, NULL);
	pan_g_ptr_array_foreach (mail_rcpts, (GFunc)g_free, NULL);
	g_ptr_array_free (mail_rcpts, TRUE);
	g_ptr_array_free (nntp_rcpts, TRUE);

	return retval;
}
